from typing import Dict, Optional
import time, requests, os

from fake_useragent import FakeUserAgent

Response = requests.Response

def download(
    url: str,
    path: str,
    max_request_try_count: int = 3,
    sleep_time: float = 2.5,
    debug: bool = False
) -> bool:
    current_try_count = 0

    while current_try_count < max_request_try_count:
        current_try_count += 1

        if debug:
            print(url + ' | ' + str(current_try_count) + '/' + str(max_request_try_count))
        
        res = __download(
            url,
            path,
            debug=debug
        )

        if res:
            return True
        
        time.sleep(sleep_time)
    
    return False

def __download(
    url: str, 
    path: str,
    debug: bool = False
) -> bool:
    import urllib

    try:
        urllib.request.urlretrieve(url, path)

        return os.path.exists(path)
    except Exception as e1:
        if debug:
            print(e1)
        
        try:
            os.remove(path)
        except Exception as e2:
            if debug:
                print(e2)
        
        return False

def request(
    url: str,
    headers: Dict = None,
    max_request_try_count: int = 10,
    sleep_time: float = 2.5,
    debug: bool = False,
    fake_useragent: bool = False
) -> Optional[Response]:
    current_try_count = 0

    while current_try_count < max_request_try_count:
        current_try_count += 1
        
        if debug:
            print(url + ' | ' + str(current_try_count) + '/' + str(max_request_try_count))

        resp = __request(
            url,
            headers=headers,
            debug=debug,
            fake_useragent=fake_useragent
        )

        if resp is not None:
            return resp
        
        time.sleep(sleep_time)
    
    return None

def __request(
    url: str,
    headers: Dict = None,
    debug: bool = False,
    fake_useragent: bool = False
) -> Optional[Response]:
    if headers is None:
        headers = {}
    
    ua = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:74.0) Gecko/20100101 Firefox/74.0'

    if fake_useragent:
        ua = FakeUserAgent().random
    
    headers = __headers_by_optionally_setting(headers, {
        'User-Agent':ua,
        'Accept':'*/*',
        'Cache-Control':'no-cache',
        'Accept-Encoding':'gzip, deflate, br',
        'Connection':'keep-alive'
    })

    try:
        resp = requests.get(url, headers=headers)

        if resp is None:
            if debug:
                print('ERROR: Resp is None')
            
            return None

        if resp.status_code != 200:
            if debug:
                print('ERROR:', resp)
            
            return None
        
        return resp
    except Exception as e:
        if debug:
            print('ERROR:', e)

        return None

def __headers_by_optionally_setting(
    headers: Dict, 
    keys_values: Dict
) -> Dict:
    for key, value in keys_values.items():
        if key not in headers:
            headers[key] = value
    
    return headers


from typing import List, Optional
import os

# allowed_extensions is an array, like ['jpg', 'jpeg', 'png']
def file_paths_from_folder(
    root_folder_path: str, 
    allowed_extensions: List[str] = None, 
    recursive: bool = True
) -> List[str]:
    import os

    root_folder_path = os.path.abspath(root_folder_path)
    file_paths = []

    for (dir_path, _, file_names) in os.walk(root_folder_path):
        abs_dir_path = os.path.abspath(dir_path)

        for file_name in file_names:
            if allowed_extensions is not None and len(allowed_extensions) > 0:
                for extension in allowed_extensions:
                    if file_name.lower().endswith(extension.lower()):
                        file_paths.append(os.path.join(abs_dir_path, file_name))

                        break
            else:
                file_paths.append(os.path.join(abs_dir_path, file_name))
        
        if not recursive:
            break
    
    return file_paths

def path_of_file(f:  str) -> str:
    return os.path.abspath(f)

def folder_path_of_file(f: str) -> str:
    return os.path.dirname(path_of_file(f))

def temp_path_for_path(_path: str) -> str:
    import random, string

    folder_path = folder_path_of_file(_path)
    ext = extension(_path, include_dot=True)

    while True:
        proposed_path = os.path.join(
            folder_path,
            '.' + ''.join(random.choices(string.ascii_lowercase + string.digits, k=8)) + ext
        )

        if not os.path.exists(proposed_path):
            return proposed_path

def file_name(_path: str, include_extension: bool = True) -> str:
    basename = os.path.basename(_path)

    if not include_extension:
        basename = remove_extensions(basename)
    
    return basename

def extension(_path: str, include_dot: bool = False) -> Optional[str]:
    path_comps = _path.replace('/.', '/').split(".")

    if len(path_comps) == 1:
        return None
    
    ext = path_comps[-1]

    if include_dot:
        ext = '.' + ext
    
    return ext

def replace_extension(_path: str, new_extension: str) -> str:
    if not new_extension.startswith('.'):
        new_extension = '.' + new_extension
    
    return _path.replace(extension(_path, include_dot=True), new_extension)

def remove_extensions(_path: str) -> str:
    while True:
        ext = extension(_path, include_dot=True)

        if ext is None:
            return _path
        
        _path = _path.rstrip(ext)

def remove(_path: str) -> bool:
    try:
        os.remove(_path)

        return True
    except:
        return False


def path(path: str) -> str:
    return path.replace(' ', '\\ ').replace('\\\\ ', '\\ ')

def sh(
    cmd: str,
    debug: bool = False
) -> str:
    import subprocess
    
    if debug:
        print(cmd)

    return subprocess.getoutput(cmd)

def pwd() -> str:
    return sh('pwd')

def mkdir(p: str) -> str:
    return sh('mkdir ' + path(p))

def rmrf(p: str) -> str:
    return sh('rm -rf ' + path(p))

def cp(from_path: str, to_path: str) -> str:
    return sh('cp ' + path(from_path) + ' ' + path(to_path))